@@ -16,11 +16,12 @@
 try:
     import ast
 except ImportError:
-    from chameleon import ast24 as ast
+    from chameleon import ast25 as ast
 
 import sys
 import logging
 import weakref
+import collections
 
 node_annotations = weakref.WeakKeyDictionary()
 
@@ -78,6 +79,45 @@ def walk_names(target, mode):
             yield node.id
 
 
+def iter_fields(node):
+    """
+    Yield a tuple of ``(fieldname, value)`` for each field in ``node._fields``
+    that is present on *node*.
+    """
+    for field in node._fields:
+        try:
+            yield field, getattr(node, field)
+        except AttributeError:
+            pass
+
+
+def iter_child_nodes(node):
+    """
+    Yield all direct child nodes of *node*, that is, all fields that are nodes
+    and all items of fields that are lists of nodes.
+    """
+    for name, field in iter_fields(node):
+        if isinstance(field, Node):
+            yield field
+        elif isinstance(field, list):
+            for item in field:
+                if isinstance(item, Node):
+                    yield item
+
+
+def walk(node):
+    """
+    Recursively yield all descendant nodes in the tree starting at *node*
+    (including *node* itself), in no specified order.  This is useful if you
+    only want to modify nodes in place and don't care about the context.
+    """
+    todo = collections.deque([node])
+    while todo:
+        node = todo.popleft()
+        todo.extend(iter_child_nodes(node))
+        yield node
+
+
 def copy(source, target):
     target.__class__ = source.__class__
     target.__dict__ = source.__dict__
@@ -119,6 +159,14 @@ class Node(object):
             id(self)
             )
 
+    def extract(self, condition):
+        result = []
+        for node in walk(self):
+            if condition(node):
+                result.append(node)
+
+        return result
+
 
 class Builtin(Node):
     """Represents a Python builtin.
@@ -270,6 +318,9 @@ class ASTCodeGenerator(object):
                 first = False
             self._write('**' + node.kwarg)
 
+    def visit_arg(self, node):
+        self._write(node.arg)
+
     # FunctionDef(identifier name, arguments args,
     #                           stmt* body, expr* decorators)
     def visit_FunctionDef(self, node):
@@ -446,6 +497,34 @@ class ASTCodeGenerator(object):
         self._write(', ')
         self.visit(node.tback)
 
+    # Try(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
+    def visit_Try(self, node):
+        self._new_line()
+        self._write('try:')
+        self._change_indent(1)
+        for statement in node.body:
+            self.visit(statement)
+        self._change_indent(-1)
+        if getattr(node, 'handlers', None):
+            for handler in node.handlers:
+                self.visit(handler)
+        self._new_line()
+
+        if getattr(node, 'orelse', None):
+            self._write('else:')
+            self._change_indent(1)
+            for statement in node.orelse:
+                self.visit(statement)
+            self._change_indent(-1)
+
+        if getattr(node, 'finalbody', None):
+            self._new_line()
+            self._write('finally:')
+            self._change_indent(1)
+            for statement in node.finalbody:
+                self.visit(statement)
+            self._change_indent(-1)
+
     # TryExcept(stmt* body, excepthandler* handlers, stmt* orelse)
     def visit_TryExcept(self, node):
         self._new_line()
@@ -835,6 +914,9 @@ class ASTCodeGenerator(object):
             self._write(', ')
         self._write(')')
 
+    # NameConstant(singleton value)
+    def visit_NameConstant(self, node):
+        self._write(str(node.value))
 
 class AnnotationAwareVisitor(ast.NodeVisitor):
     def visit(self, node):
@@ -870,6 +952,13 @@ class NameLookupRewriteVisitor(AnnotationAwareVisitor):
             self.transformed.add(node.id)
             self.apply_transform(node)
 
+    def visit_FunctionDef(self, node):
+        self.scopes[-1].add(node.name)
+
+    def visit_alias(self, node):
+        name = node.asname if node.asname is not None else node.name
+        self.scopes[-1].add(name)
+
     def visit_Lambda(self, node):
         self.scopes.append(set())
         try:
