@@ -1,4 +1,5 @@
 import re
+import cgi
 import sys
 import itertools
 import logging
@@ -6,6 +7,7 @@ import threading
 import functools
 import collections
 import pickle
+import textwrap
 
 from .astutil import load
 from .astutil import store
@@ -18,13 +20,14 @@ from .astutil import NameLookupRewriteVisitor
 from .astutil import Comment
 from .astutil import Symbol
 from .astutil import Builtin
+from .astutil import Static
 
 from .codegen import TemplateCodeGenerator
 from .codegen import template
 
 from .tal import ErrorInfo
 from .tal import NAME
-from .i18n import fast_translate
+from .i18n import simple_translate
 
 from .nodes import Text
 from .nodes import Value
@@ -73,7 +76,7 @@ COMPILER_INTERNALS_OR_DISALLOWED = set([
     ])
 
 
-RE_MANGLE = re.compile('[\-: ]')
+RE_MANGLE = re.compile('[^\w_]')
 RE_NAME = re.compile('^%s$' % NAME)
 
 if DEBUG_MODE:
@@ -87,7 +90,7 @@ def identifier(prefix, suffix=None):
 
 
 def mangle(string):
-    return RE_MANGLE.sub('_', str(string)).replace('\n', '')
+    return RE_MANGLE.sub('_', str(string)).replace('\n', '').replace('-', '_')
 
 
 def load_econtext(name):
@@ -187,65 +190,96 @@ def emit_convert(
                 target = target()
 
 
+@template
+def emit_func_convert(
+    func, encoded=byte_string, str=unicode_string,
+    long=long, type=type):  # pragma: no cover
+    def func(target):
+        if target is None:
+            return
+
+        __tt = type(target)
+
+        if __tt is int or __tt is float or __tt is long:
+            target = str(target)
+
+        elif __tt is encoded:
+            target = decode(target)
+
+        elif __tt is not str:
+            try:
+                target = target.__html__
+            except AttributeError:
+                __converted = convert(target)
+                target = str(target) if target is __converted else __converted
+            else:
+                target = target()
+
+        return target
+
+
 @template
 def emit_translate(target, msgid, default=None):  # pragma: no cover
     target = translate(msgid, default=default, domain=__i18n_domain)
 
 
 @template
-def emit_convert_and_escape(
-    target, quote=None, quote_entity=None, str=unicode_string, long=long,
-    type=type, encoded=byte_string,
-    default_marker=None, default=None):  # pragma: no cover
-    if target is None:
-        pass
-    elif target is default_marker:
-        target = default
-    else:
+def emit_func_convert_and_escape(
+    func, str=unicode_string, long=long,
+    type=type, encoded=byte_string):  # pragma: no cover
+
+    def func(target, quote, quote_entity, default, default_marker):
+        if target is None:
+            return
+
+        if target is default_marker:
+            return default
+
         __tt = type(target)
 
         if __tt is int or __tt is float or __tt is long:
             target = str(target)
         else:
-            try:
-                if __tt is encoded:
-                    target = decode(target)
-                elif __tt is not str:
-                    try:
-                        target = target.__html__
-                    except:
-                        __converted = convert(target)
-                        target = str(target) if target is __converted \
-                                 else __converted
-                    else:
-                        raise RuntimeError
-            except RuntimeError:
-                target = target()
-            else:
-                if target is not None:
-                    try:
-                        escape = __re_needs_escape(target) is not None
-                    except TypeError:
-                        pass
-                    else:
-                        if escape:
-                            # Character escape
-                            if '&' in target:
-                                target = target.replace('&', '&amp;')
-                            if '<' in target:
-                                target = target.replace('<', '&lt;')
-                            if '>' in target:
-                                target = target.replace('>', '&gt;')
-                            if quote is not None and quote in target:
-                                target = target.replace(quote, quote_entity)
+            if __tt is encoded:
+                target = decode(target)
+            elif __tt is not str:
+                try:
+                    target = target.__html__
+                except:
+                    __converted = convert(target)
+                    target = str(target) if target is __converted \
+                             else __converted
+                else:
+                    return target()
+
+            if target is not None:
+                try:
+                    escape = __re_needs_escape(target) is not None
+                except TypeError:
+                    pass
+                else:
+                    if escape:
+                        # Character escape
+                        if '&' in target:
+                            target = target.replace('&', '&amp;')
+                        if '<' in target:
+                            target = target.replace('<', '&lt;')
+                        if '>' in target:
+                            target = target.replace('>', '&gt;')
+                        if quote is not None and quote in target:
+                            target = target.replace(quote, quote_entity)
+
+        return target
 
 
 class Interpolator(object):
     braces_required_regex = re.compile(
-        r'(?<!\\)\$({(?P<expression>.*)})')
+        r'(?<!\\)\$({(?P<expression>.*)})',
+        re.DOTALL)
 
     braces_optional_regex = re.compile(
-        r'(?<!\\)\$({(?P<expression>.*)}|(?P<variable>[A-Za-z][A-Za-z0-9_]*))')
+        r'(?<!\\)\$({(?P<expression>.*)}|(?P<variable>[A-Za-z][A-Za-z0-9_]*))',
+        re.DOTALL)
 
     def __init__(self, expression, braces_required, translate=False):
         self.expression = expression
@@ -306,20 +340,26 @@ class Interpolator(object):
 
             while True:
                 d = groupdict(m, matched)
-                string = d["expression"] or d["variable"] or ""
-
+                string = d["expression"] or d.get("variable") or ""
                 string = decode_htmlentities(string)
 
-                try:
-                    compiler = engine.parse(string)
-                    body += compiler.assign_text(target)
-                except ExpressionError:
-                    matched = matched[m.start():m.end() - 1]
-                    m = self.regex.search(matched)
-                    if m is None:
-                        raise
+                if string:
+                    try:
+                        compiler = engine.parse(string)
+                        body += compiler.assign_text(target)
+                    except ExpressionError:
+                        matched = matched[m.start():m.end() - 1]
+                        m = self.regex.search(matched)
+                        if m is None:
+                            raise
+
+                        continue
                 else:
-                    break
+                    s = m.group()
+                    assign = ast.Assign(targets=[target], value=ast.Str(s=s))
+                    body += [assign]
+
+                break
 
             # If one or more expressions are not simple names, we
             # disable translation.
@@ -341,7 +381,7 @@ class Interpolator(object):
 
             if translate and isinstance(target, ast.Str):
                 target = template(
-                    "translate(msgid, domain=__i18n_domain)",
+                    "translate(msgid, domain=__i18n_domain, context=econtext)",
                     msgid=target, mode="eval",
                     )
         else:
@@ -360,7 +400,7 @@ class Interpolator(object):
                         values.append(node)
 
                 target = template(
-                    "translate(msgid, mapping=mapping, domain=__i18n_domain)",
+                    "translate(msgid, mapping=mapping, domain=__i18n_domain, context=econtext)",
                     msgid=ast.Str(s=formatting_string),
                     mapping=ast.Dict(keys=keys, values=values),
                     mode="eval"
@@ -496,12 +536,13 @@ class ExpressionEngine(object):
 
             entity = char2entity(quote or '\0')
 
-            return emit_convert_and_escape(
-                target,
-                quote=ast.Str(s=quote),
-                quote_entity=ast.Str(s=entity),
-                default=self._default,
-                default_marker=self._default_marker,
+            return template(
+                "TARGET = __quote(TARGET, QUOTE, Q_ENTITY, DEFAULT, MARKER)",
+                TARGET=target,
+                QUOTE=ast.Str(s=quote),
+                Q_ENTITY=ast.Str(s=entity),
+                DEFAULT=self._default,
+                MARKER=self._default_marker,
                 )
 
         return emit_convert(
@@ -683,11 +724,11 @@ class ExpressionTransform(object):
 
     loads_symbol = Symbol(pickle.loads)
 
-    def __init__(self, engine_factory, cache, transform, strict=True):
+    def __init__(self, engine_factory, cache, visitor, strict=True):
         self.engine_factory = engine_factory
         self.cache = cache
-        self.transform = transform
         self.strict = strict
+        self.visitor = visitor
 
     def __call__(self, expression, target):
         if isinstance(target, string_type):
@@ -711,11 +752,9 @@ class ExpressionTransform(object):
             stmts += set_error(token, load("__exc"))
             stmts += [ast.Raise(exc=load("__exc"))]
 
-        # Apply dynamic name rewrite transform to each statement
-        visitor = NameLookupRewriteVisitor(self.transform)
-
+        # Apply visitor to each statement
         for stmt in stmts:
-            visitor(stmt)
+            self.visitor(stmt)
 
         return stmts
 
@@ -751,6 +790,9 @@ class ExpressionTransform(object):
         compiler = engine.parse(node.value)
         return compiler.assign_value(target)
 
+    def visit_Copy(self, node, target):
+        return self.translate(node.expression, target)
+
     def visit_Default(self, node, target):
         value = annotated(node.marker)
         return [ast.Assign(targets=[target], value=value)]
@@ -836,7 +878,7 @@ class Compiler(object):
                  TypeError
 
     defaults = {
-        'translate': Symbol(fast_translate),
+        'translate': Symbol(simple_translate),
         'decode': Builtin("str"),
         'convert': Builtin("str"),
         }
@@ -863,10 +905,12 @@ class Compiler(object):
             internals,
             )
 
+        self._visitor = visitor = NameLookupRewriteVisitor(transform)
+
         self._engine = ExpressionTransform(
             engine_factory,
             self._expression_cache,
-            transform,
+            visitor,
             strict=strict,
             )
 
@@ -903,8 +947,6 @@ class Compiler(object):
                 yield stmt
 
     def visit_Element(self, node):
-        self._aliases.append(self._aliases[-1].copy())
-
         for stmt in self.visit(node.start):
             yield stmt
 
@@ -915,13 +957,12 @@ class Compiler(object):
             for stmt in self.visit(node.end):
                 yield stmt
 
-        self._aliases.pop()
-
     def visit_Module(self, node):
         body = []
 
         body += template("import re")
         body += template("import functools")
+        body += template("from itertools import chain as __chain")
         body += template("__marker = object()")
         body += template(
             r"g_re_amp = re.compile(r'&(?!([A-Za-z]+|#[0-9]+);)')"
@@ -980,6 +1021,9 @@ class Compiler(object):
         body += template("__re_amp = g_re_amp")
         body += template("__re_needs_escape = g_re_needs_escape")
 
+        body += emit_func_convert("__convert")
+        body += emit_func_convert_and_escape("__quote")
+
         # Resolve defaults
         for name in self.defaults:
             body += template(
@@ -1025,7 +1069,7 @@ class Compiler(object):
         return emit_node(ast.Str(s=node.value))
 
     def visit_Domain(self, node):
-        backup = "__previous_i18n_domain_%d" % id(node)
+        backup = "__previous_i18n_domain_%s" % mangle(id(node))
         return template("BACKUP = __i18n_domain", BACKUP=backup) + \
                template("__i18n_domain = NAME", NAME=ast.Str(s=node.name)) + \
                self.visit(node.node) + \
@@ -1069,9 +1113,12 @@ class Compiler(object):
             body += emit_translate(name, name)
 
         if node.char_escape:
-            body += emit_convert_and_escape(name)
+            body += template(
+                "NAME=__quote(NAME, None, '\255', None, None)",
+                NAME=name,
+                )
         else:
-            body += emit_convert(name)
+            body += template("NAME = __convert(NAME)", NAME=name)
 
         body += template("if NAME is not None: __append(NAME)", NAME=name)
 
@@ -1124,6 +1171,7 @@ class Compiler(object):
     def visit_Define(self, node):
         scope = set(self._scopes[-1])
         self._scopes.append(scope)
+        self._aliases.append(self._aliases[-1].copy())
 
         for assignment in node.assignments:
             if assignment.local:
@@ -1142,6 +1190,7 @@ class Compiler(object):
                     yield stmt
 
         self._scopes.pop()
+        self._aliases.pop()
 
     def visit_Omit(self, node):
         return self.visit_Condition(node)
@@ -1226,8 +1275,8 @@ class Compiler(object):
 
         # emit the translation expression
         body += template(
-            "__append(translate("
-            "msgid, mapping=mapping, default=default, domain=__i18n_domain))",
+            "if msgid: __append(translate("
+            "msgid, mapping=mapping, default=default, domain=__i18n_domain, context=econtext))",
             msgid=msgid, default=default, mapping=mapping
             )
 
@@ -1251,9 +1300,8 @@ class Compiler(object):
             for stmt in emit_node(ast.Str(s=node.prefix + node.name)):
                 yield stmt
 
-            for attribute in node.attributes:
-                for stmt in self.visit(attribute):
-                    yield stmt
+            for stmt in self.visit(node.attributes):
+                yield stmt
 
             for stmt in emit_node(ast.Str(s=node.suffix)):
                 yield stmt
@@ -1268,21 +1316,78 @@ class Compiler(object):
             yield stmt
 
     def visit_Attribute(self, node):
-        f = node.space + node.name + node.eq + node.quote + "%s" + node.quote
+        attr_format = (node.space + node.name + node.eq +
+                       node.quote + "%s" + node.quote)
+
+        filter_args = list(map(self._engine.cache.get, node.filters))
+
+        filter_condition = template(
+            "NAME not in CHAIN",
+            NAME=ast.Str(s=node.name),
+            CHAIN=ast.Call(
+                func=load("__chain"),
+                args=filter_args,
+                keywords=[],
+                starargs=None,
+                kwargs=None,
+            ),
+            mode="eval"
+        )
 
         # Static attributes are just outputted directly
         if isinstance(node.expression, ast.Str):
-            s = f % node.expression.s
-            return template("__append(S)", S=ast.Str(s=s))
+            s = attr_format % node.expression.s
+            if node.filters:
+                return template(
+                    "if C: __append(S)", C=filter_condition, S=ast.Str(s=s)
+                )
+            else:
+                return template("__append(S)", S=ast.Str(s=s))
 
         target = identifier("attr", node.name)
         body = self._engine(node.expression, store(target))
-        return body + template(
-            "if TARGET is not None: __append(FORMAT % TARGET)",
-            FORMAT=ast.Str(s=f),
-            TARGET=target,
+
+        condition = template("TARGET is not None", TARGET=target, mode="eval")
+
+        if node.filters:
+            condition = ast.BoolOp(
+                values=[condition, filter_condition],
+                op=ast.And(),
             )
 
+        return body + template(
+            "if CONDITION: __append(FORMAT % TARGET)",
+            FORMAT=ast.Str(s=attr_format),
+            TARGET=target,
+            CONDITION=condition,
+        )
+
+    def visit_DictAttributes(self, node):
+        target = identifier("attr", id(node))
+        body = self._engine(node.expression, store(target))
+
+        exclude = Static(template(
+            "set(LIST)", LIST=ast.List(
+                elts=[ast.Str(s=name) for name in node.exclude],
+                ctx=ast.Load(),
+            ), mode="eval"
+        ))
+
+        body += template(
+            "for name, value in TARGET.items():\n  "
+            "if name not in EXCLUDE and value is not None: __append("
+            "' ' + name + '=' + QUOTE + "
+            "QUOTE_FUNC(value, QUOTE, QUOTE_ENTITY, None, None) + QUOTE"
+            ")",
+            TARGET=target,
+            EXCLUDE=exclude,
+            QUOTE_FUNC="__quote",
+            QUOTE=ast.Str(s=node.quote),
+            QUOTE_ENTITY=ast.Str(s=char2entity(node.quote or '\0')),
+            )
+
+        return body
+
     def visit_Cache(self, node):
         body = []
 
@@ -1301,6 +1406,22 @@ class Compiler(object):
 
         return body
 
+    def visit_Cancel(self, node):
+        body = []
+
+        for expression in node.expressions:
+            name = identifier("cache", id(expression))
+            target = store(name)
+
+            if not self._expression_cache.get(expression):
+               continue
+
+            body.append(ast.Assign([target], load("None")))
+
+        body += self.visit(node.node)
+
+        return body
+
     def visit_UseInternalMacro(self, node):
         if node.name is None:
             render = "render"
@@ -1319,7 +1440,7 @@ class Compiler(object):
         self._slots.add(name)
 
         orelse = template(
-            "SLOT(__stream, econtext.copy(), rcontext, __i18n_domain)",
+            "SLOT(__stream, econtext.copy(), rcontext)",
             SLOT=name)
         test = ast.Compare(
             left=load(name),
@@ -1340,7 +1461,7 @@ class Compiler(object):
 
         if node.name in self._translations[-1]:
             raise TranslationError(
-                "Duplicate translation name: %s." % node.name)
+                "Duplicate translation name: %s.", node.name)
 
         self._translations[-1].add(node.name)
         body = []
@@ -1364,6 +1485,14 @@ class Compiler(object):
 
         return body
 
+    def visit_CodeBlock(self, node):
+        stmts = template(textwrap.dedent(node.source.strip('\n')))
+
+        for stmt in stmts:
+            self._visitor(stmt)
+
+        return [try_except_wrap(stmts, node.source)]
+
     def visit_UseExternalMacro(self, node):
         self._macros.append(node.extend)
 
@@ -1517,9 +1646,9 @@ class Compiler(object):
 
     def _get_translation_identifiers(self, name):
         assert self._translations
-        prefix = id(self._translations[-1])
-        stream = identifier("stream_%d" % prefix, name)
-        append = identifier("append_%d" % prefix, name)
+        prefix = str(id(self._translations[-1])).replace('-', '_')
+        stream = identifier("stream_%s" % prefix, name)
+        append = identifier("append_%s" % prefix, name)
         return stream, append
 
     def _enter_assignment(self, names):
