@@ -168,7 +168,7 @@ class ZopePageTemplatesTest(RenderTestCase):
                     exc = sys.exc_info()[1]
                     return func(self, body, exc)
                 else:
-                    self.fail("Expected exception; got: %s." % result)
+                    self.fail("Expected exception.")
 
             return wrapper
         return decorator
@@ -201,6 +201,26 @@ class ZopePageTemplatesTest(RenderTestCase):
     def test_attributes_on_tal_tag_fails(self, body, exc):
         self.assertTrue(body[exc.offset:].startswith('dummy'))
 
+    @error("""<tal:dummy i18n:attributes=\"foo, bar\" />""")
+    def test_i18n_attributes_with_non_identifiers(self, body, exc):
+        self.assertTrue(body[exc.offset:].startswith('foo,'))
+
+    @error("""<tal:dummy repeat=\"key,value mydict.items()\">""")
+    def test_repeat_syntax_error_message(self, body, exc):
+        self.assertTrue(body[exc.offset:].startswith('key,value'))
+
+    @error('''<tal:dummy><p i18n:translate="mymsgid">
+            <span i18n:name="repeat"/><span i18n:name="repeat"/>
+            </p></tal:dummy>''')
+    def test_repeat_i18n_name_error(self, body, exc):
+        self.assertTrue(body[exc.offset:].startswith('repeat'), body[exc.offset:])
+
+    @error('''<tal:dummy>
+            <span i18n:name="not_in_translation"/>
+            </tal:dummy>''')
+    def test_i18n_name_not_in_translation_error(self, body, exc):
+        self.assertTrue(body[exc.offset:].startswith('not_in_translation'))
+
     def test_encoded(self):
         filename = '074-encoded-template.pt'
         with open(os.path.join(self.root, 'inputs', filename), 'rb') as f:
@@ -323,6 +343,24 @@ class ZopePageTemplatesTest(RenderTestCase):
             )
         self.assertTrue(template(), "<div>foo</div>")
 
+    def test_trim_attribute_space(self):
+        document = '''<div
+                  class="document"
+                  id="test"
+                  tal:attributes="class string:${default} test"
+            />'''
+
+        result1 = self.from_string(
+            document)()
+
+        result2 = self.from_string(
+            document, trim_attribute_space=True)()
+
+        self.assertEqual(result1.count(" "), 49)
+        self.assertEqual(result2.count(" "), 4)
+        self.assertTrue(" />" in result1)
+        self.assertTrue(" />" in result2)
+
     def test_exception(self):
         from traceback import format_exception_only
 
@@ -336,13 +374,67 @@ class ZopePageTemplatesTest(RenderTestCase):
             formatted = str(exc)
             self.assertFalse('NameError:' in formatted)
             self.assertTrue('foo' in formatted)
-            self.assertTrue('(1:23)' in formatted)
+            self.assertTrue('(line 1: col 23)' in formatted)
 
             formatted_exc = "\n".join(format_exception_only(type(exc), exc))
             self.assertTrue('NameError: foo' in formatted_exc)
         else:
             self.fail("expected error")
 
+    def test_create_formatted_exception(self):
+        from chameleon.utils import create_formatted_exception
+
+        exc = create_formatted_exception(NameError('foo'), NameError, str)
+        self.assertEqual(exc.args, ('foo', ))
+
+        class MyNameError(NameError):
+            def __init__(self, boo):
+                NameError.__init__(self, boo)
+                self.bar = boo
+
+        exc = create_formatted_exception(MyNameError('foo'), MyNameError, str)
+        self.assertEqual(exc.args, ('foo', ))
+        self.assertEqual(exc.bar, 'foo')
+
+    def test_create_formatted_exception_no_subclass(self):
+        from chameleon.utils import create_formatted_exception
+
+        class DifficultMetaClass(type):
+            def __init__(self, class_name, bases, namespace):
+                if not bases == (BaseException, ):
+                    raise TypeError(bases)
+
+        Difficult = DifficultMetaClass('Difficult', (BaseException, ), {'args': ()})
+
+        exc = create_formatted_exception(Difficult(), Difficult, str)
+        self.assertEqual(exc.args, ())
+
+    def test_error_handler_makes_safe_copy(self):
+        calls = []
+
+        class TestException(Exception):
+            def __init__(self, *args, **kwargs):
+                calls.append((args, kwargs))
+
+        def _render(stream, econtext, rcontext):
+            exc = TestException('foo', bar='baz')
+            rcontext['__error__'] = ('expression', 1, 42, 'test.pt', exc),
+            raise exc
+
+        template = self.from_string("")
+        template._render = _render
+        try:
+            template()
+        except TestException:
+            self.assertEqual(calls, [(('foo', ), {'bar': 'baz'})])
+            exc = sys.exc_info()[1]
+            formatted = str(exc)
+            self.assertTrue('TestException' in formatted)
+            self.assertTrue('"expression"' in formatted)
+            self.assertTrue('(line 1: col 42)' in formatted)
+        else:
+            self.fail("unexpected error")
+
     def test_double_underscore_variable(self):
         from chameleon.exc import TranslationError
         self.assertRaises(
@@ -358,6 +450,14 @@ class ZopePageTemplatesTest(RenderTestCase):
             body = "<d tal:define=\"%s 'foo'\">${%s}</d>" % (name, name)
             self.assertRaises(TranslationError, self.from_string, body)
 
+    def test_simple_translate_mapping(self):
+        template = self.from_string(
+            '<div i18n:translate="">'
+            '<span i18n:name="name">foo</span>'
+            '</div>')
+
+        self.assertEqual(template(), '<div><span>foo</span></div>')
+
     def test_translate_is_not_an_internal(self):
         macro = self.from_string('<span i18n:translate="">bar</span>')
         template = self.from_string(
@@ -525,8 +625,8 @@ class ZopeTemplatesTestSuite(RenderTestCase):
             # source when debugging
             self.shortDescription = lambda: input_path
 
-            # Very implicitly enable implicit translation based on
-            # a string included in the input path:
+            # When input path contaiins the string 'implicit-i18n', we
+            # enable "implicit translation".
             implicit_i18n = 'implicit-i18n' in input_path
             implicit_i18n_attrs = ("alt", "title") if implicit_i18n else ()
 
