@@ -3,7 +3,7 @@ import re
 try:
     import ast
 except ImportError:
-    from chameleon import ast24 as ast
+    from chameleon import ast25 as ast
 
 try:
     str = unicode
@@ -11,6 +11,7 @@ except NameError:
     long = int
 
 from functools import partial
+from copy import copy
 
 from ..program import ElementProgram
 
@@ -44,6 +45,10 @@ except NameError:
 
 missing = object()
 
+re_trim = re.compile(r'($\s+|\s+^)', re.MULTILINE)
+
+EMPTY_DICT = Static(ast.Dict(keys=[], values=[]))
+
 
 def skip(node):
     return node
@@ -110,6 +115,9 @@ class MacroProgram(ElementProgram):
     # If set, text will be translated even without explicit markup.
     implicit_i18n_translate = False
 
+    # If set, additional attribute whitespace will be stripped.
+    trim_attribute_space = False
+
     def __init__(self, *args, **kwargs):
         # Internal array for switch statements
         self._switches = []
@@ -131,6 +139,7 @@ class MacroProgram(ElementProgram):
             'escape',
             'implicit_i18n_translate',
             'implicit_i18n_attributes',
+            'trim_attribute_space',
             )
 
         super(MacroProgram, self).__init__(*args, **kwargs)
@@ -176,7 +185,8 @@ class MacroProgram(ElementProgram):
         except KeyError:
             switch = None
         else:
-            switch = nodes.Value(clause)
+            value = nodes.Value(clause)
+            switch = value, nodes.Copy(value)
 
         self._switches.append(switch)
 
@@ -186,6 +196,7 @@ class MacroProgram(ElementProgram):
         use_macro = ns.get((METAL, 'use-macro'))
         extend_macro = ns.get((METAL, 'extend-macro'))
         if use_macro or extend_macro:
+            omit = True
             slots = []
             self._use_macro.append(slots)
 
@@ -238,7 +249,7 @@ class MacroProgram(ElementProgram):
             try:
                 clause = ns[TAL, 'attributes']
             except KeyError:
-                TAL_ATTRIBUTES = {}
+                TAL_ATTRIBUTES = []
             else:
                 TAL_ATTRIBUTES = tal.parse_attributes(clause)
 
@@ -259,22 +270,22 @@ class MacroProgram(ElementProgram):
             # Create attribute nodes
             STATIC_ATTRIBUTES = self._create_static_attributes(prepared)
             ATTRIBUTES = self._create_attributes_nodes(
-                prepared, I18N_ATTRIBUTES
+                prepared, I18N_ATTRIBUTES, STATIC_ATTRIBUTES
                 )
 
             # Start- and end nodes
             start_tag = nodes.Start(
                 start['name'],
-                start['prefix'],
-                start['suffix'],
+                self._maybe_trim(start['prefix']),
+                self._maybe_trim(start['suffix']),
                 ATTRIBUTES
                 )
 
             end_tag = nodes.End(
                 end['name'],
                 end['space'],
-                end['prefix'],
-                end['suffix'],
+                self._maybe_trim(end['prefix']),
+                self._maybe_trim(end['suffix']),
                 ) if end is not None else None
 
             # tal:omit-tag
@@ -308,7 +319,7 @@ class MacroProgram(ElementProgram):
 
                 # Assign static attributes dictionary to "attrs" value
                 inner = nodes.Define(
-                    [nodes.Alias(["attrs"], STATIC_ATTRIBUTES)],
+                    [nodes.Alias(["attrs"], STATIC_ATTRIBUTES or EMPTY_DICT)],
                     inner,
                     )
 
@@ -366,12 +377,11 @@ class MacroProgram(ElementProgram):
                     )
 
             CASE = lambda node: nodes.Define(
-                [nodes.Assignment(["default"], switch, True)],
+                [nodes.Alias(["default"], switch[1], False)],
                 nodes.Condition(
-                    nodes.Equality(switch, value),
-                    node,
-                    )
-                )
+                    nodes.Equality(switch[0], value),
+                    nodes.Cancel([switch[0]], node),
+                ))
 
         # tal:repeat
         try:
@@ -385,6 +395,11 @@ class MacroProgram(ElementProgram):
 
             expression = nodes.Value(expr)
 
+            if start['namespace'] == TAL:
+                self._last = None
+                self._whitespace = whitespace.lstrip('\n')
+                whitespace = ""
+
             REPEAT = partial(
                 nodes.Repeat,
                 names,
@@ -406,7 +421,7 @@ class MacroProgram(ElementProgram):
         if switch is None:
             SWITCH = skip
         else:
-            SWITCH = partial(nodes.Cache, [switch])
+            SWITCH = partial(nodes.Cache, list(switch))
 
         # i18n:domain
         try:
@@ -422,7 +437,10 @@ class MacroProgram(ElementProgram):
         except KeyError:
             NAME = skip
         else:
-            NAME = partial(nodes.Name, clause)
+            if not clause.strip():
+                NAME = skip
+            else:
+                NAME = partial(nodes.Name, clause)
 
         # The "slot" node next is the first node level that can serve
         # as a macro slot
@@ -443,6 +461,12 @@ class MacroProgram(ElementProgram):
         except KeyError:
             pass
         else:
+            if not clause.strip():
+                raise LanguageError(
+                    "Must provide a non-trivial string for metal:fill-slot.",
+                    clause
+                )
+
             index = -(1 + int(bool(use_macro or extend_macro)))
 
             try:
@@ -478,11 +502,34 @@ class MacroProgram(ElementProgram):
         else:
             key, value = tal.parse_substitution(clause)
             translate = True if ns.get((I18N, 'translate')) == '' else False
-            fallback = nodes.Element(
-                start_tag,
-                end_tag,
-                self._make_content_node(value, None, key, translate)
+
+            fallback = self._make_content_node(value, None, key, translate)
+
+            if omit is False and start['namespace'] not in self.DROP_NS:
+                start_tag = copy(start_tag)
+
+                start_tag.attributes = nodes.Sequence(
+                    start_tag.attributes.extract(
+                        lambda attribute:
+                        isinstance(attribute, nodes.Attribute) and
+                        isinstance(attribute.expression, ast.Str)
+                    )
                 )
+
+                if end_tag is None:
+                    # Make sure start-tag has opening suffix. We don't
+                    # allow self-closing element here.
+                    start_tag.suffix  = ">"
+
+                    # Explicitly set end-tag.
+                    end_tag = nodes.End(start_tag.name, '', '</', '>',)
+
+                fallback = nodes.Element(
+                    start_tag,
+                    end_tag,
+                    fallback,
+                )
+
             ON_ERROR = partial(nodes.OnError, fallback, 'error')
 
         clause = ns.get((META, 'interpolation'))
@@ -515,6 +562,13 @@ class MacroProgram(ElementProgram):
     def visit_start_tag(self, start):
         return self.visit_element(start, None, [])
 
+    def visit_cdata(self, node):
+        if not self._interpolation[-1] or not '${' in node:
+            return nodes.Text(node)
+
+        expr = nodes.Substitution(node, ())
+        return nodes.Interpolation(expr, True, False)
+
     def visit_comment(self, node):
         if node.startswith('<!--!'):
             return
@@ -534,6 +588,13 @@ class MacroProgram(ElementProgram):
              nodes.Text(node[-3:])
              ])
 
+    def visit_processing_instruction(self, node):
+        if node['name'] != 'python':
+            text = '<?' + node['name'] + node['text'] + '?>'
+            return self.visit_text(text)
+
+        return nodes.CodeBlock(node['text'])
+
     def visit_text(self, node):
         self._last = node
 
@@ -547,29 +608,17 @@ class MacroProgram(ElementProgram):
         if not translation:
             return nodes.Text(node)
 
-        seq = []
-        while node:
-            m = re.search('\s+', node)
-            if m is not None:
-                s = m.start()
-                if s:
-                    t = node[:s]
-                    seq.append(nodes.Translate(t, nodes.Text(t)))
-
-                e = m.end()
-                whitespace = nodes.Text(node[s:e])
-                seq.append(whitespace)
-
-                if e < len(node):
-                    node = node[e:]
-                    continue
-
-            else:
-                seq.append(nodes.Translate(node, nodes.Text(node)))
-
-            break
-
-        return nodes.Sequence(seq)
+        match = re.search(r'(\s*)(.*\S)(\s*)', node, flags=re.DOTALL)
+        if match is not None:
+            prefix, text, suffix = match.groups()
+            normalized = re.sub('\s+', ' ', text)
+            return nodes.Sequence([
+                nodes.Text(prefix),
+                nodes.Translate(normalized, nodes.Text(normalized)),
+                nodes.Text(suffix),
+            ])
+        else:
+            return nodes.Text(node)
 
     def _pop_defaults(self, kwargs, *attributes):
         for attribute in attributes:
@@ -626,11 +675,17 @@ class MacroProgram(ElementProgram):
 
         return content
 
-    def _create_attributes_nodes(self, prepared, I18N_ATTRIBUTES):
+    def _create_attributes_nodes(self, prepared, I18N_ATTRIBUTES, STATIC):
         attributes = []
 
-        for name, text, quote, space, eq, expr in prepared:
-            implicit_i18n = name.lower() in self.implicit_i18n_attributes
+        names = [attr[0] for attr in prepared]
+        filtering = [[]]
+
+        for i, (name, text, quote, space, eq, expr) in enumerate(prepared):
+            implicit_i18n = (
+                name is not None and
+                name.lower() in self.implicit_i18n_attributes
+            )
 
             char_escape = ('&', '<', '>', quote)
 
@@ -656,7 +711,16 @@ class MacroProgram(ElementProgram):
             # If the expression is non-trivial, the attribute is
             # dynamic (computed).
             elif expr is not None:
-                if name in self.boolean_attributes:
+                if name is None:
+                    expression = nodes.Value(expr)
+                    value = nodes.DictAttributes(
+                        expression, ('&', '<', '>', '"'), '"',
+                        set(filter(None, names[i:]))
+                    )
+                    for fs in filtering:
+                        fs.append(expression)
+                    filtering.append([])
+                elif name in self.boolean_attributes:
                     value = nodes.Boolean(expr, name)
                 else:
                     if text is not None:
@@ -666,33 +730,60 @@ class MacroProgram(ElementProgram):
 
                     value = nodes.Substitution(expr, char_escape, default)
 
-            # Otherwise, it's a static attribute.
+            # Otherwise, it's a static attribute. We don't include it
+            # here if there's one or more "computed" attributes
+            # (dynamic, from one or more dict values).
             else:
                 value = ast.Str(s=text)
                 if msgid is missing and implicit_i18n:
                     msgid = text
 
-            # If translation is required, wrap in a translation
-            # clause
-            if msgid is not missing:
-                value = nodes.Translate(msgid, value)
+            if name is not None:
+                # If translation is required, wrap in a translation
+                # clause
+                if msgid is not missing:
+                    value = nodes.Translate(msgid, value)
 
-            attribute = nodes.Attribute(name, value, quote, eq, space)
+                space = self._maybe_trim(space)
+                fs = filtering[-1]
+                attribute = nodes.Attribute(name, value, quote, eq, space, fs)
 
-            # Always define a ``default`` alias
-            attribute = nodes.Define(
-                [nodes.Alias(["default"], default_marker)],
-                attribute,
-                )
+                if not isinstance(value, ast.Str):
+                    # Always define a ``default`` alias for non-static
+                    # expressions.
+                    attribute = nodes.Define(
+                        [nodes.Alias(["default"], default_marker)],
+                        attribute,
+                        )
 
-            attributes.append(attribute)
+                value = attribute
 
-        return attributes
+            attributes.append(value)
+
+        result = nodes.Sequence(attributes)
+
+        fs = filtering[0]
+        if fs:
+            return nodes.Cache(fs, result)
+
+        return result
 
     def _create_static_attributes(self, prepared):
         static_attrs = {}
 
         for name, text, quote, space, eq, expr in prepared:
+            if name is None:
+                continue
+
             static_attrs[name] = text if text is not None else expr
 
+        if not static_attrs:
+            return
+
         return Static(parse(repr(static_attrs)).body)
+
+    def _maybe_trim(self, string):
+        if self.trim_attribute_space:
+            return re_trim.sub(" ", string)
+
+        return string
